import sklearn
import numpy as np
import matplotlib.pyplot as plt

import torch

import copy

from . import torch_helpers, indexing


def cluster_similarity_matrices(
    s, 
    l, 
    verbose=True
):
    """
    Compute the similarity matrices for each cluster in l.
    This algorithm works best on large and sparse matrices. 
    RH 2023

    Args:
        s (scipy.sparse.csr_matrix or np.ndarray or sparse.COO):
            Similarity matrix.
            Entries should be non-negative floats.
        l (np.ndarray):
            Labels for each row of s.
            Labels should be integers ideally.
        verbose (bool):
            Whether to print warnings.

    Returns:
        cs_mean (np.ndarray):
            Similarity matrix for each cluster.
            Each element is the mean similarity between all the pairs
             of samples in each cluster.
            Note that the diagonal here only considers non-self similarity,
             which excludes the diagonals of s.
        cs_max (np.ndarray):
            Similarity matrix for each cluster.
            Each element is the maximum similarity between all the pairs
             of samples in each cluster.
            Note that the diagonal here only considers non-self similarity,
             which excludes the diagonals of s.
        cs_min (np.ndarray):
            Similarity matrix for each cluster.
            Each element is the minimum similarity between all the pairs
             of samples in each cluster. Will be 0 if there are any sparse
             elements between the two clusters.
    """
    import sparse
    import scipy.sparse

    l_arr = np.array(l)
    ss = scipy.sparse.csr_matrix(s.astype(np.float32))

    ## assert that all labels have at least two samples
    l_u ,l_c = np.unique(l_arr, return_counts=True)
    # assert np.all(l_c >= 2), "All labels must have at least two samples."
    ## assert that s is a square matrix
    assert ss.shape[0] == ss.shape[1], "Similarity matrix must be square."
    ## assert that s is non-negative
    assert (ss < 0).sum() == 0, "Similarity matrix must be non-negative."
    ## assert that l is a 1-D array
    assert len(l.shape) == 1, "Labels must be a 1-D array."
    ## assert that l is the same length as s
    assert len(l) == ss.shape[0], "Labels must be the same length as the similarity matrix."
    if verbose:
        ## Warn if s is not symmetric
        if not (ss - ss.T).sum() == 0:
            print("Warning: Similarity matrix is not symmetric.") if verbose else None
        ## Warn if s is not sparse
        if not isinstance(ss, (np.ndarray, sparse.COO, scipy.sparse.csr_matrix)):
            print("Warning: Similarity matrix is not a recognized sparse type or np.ndarray. Will attempt to convert to sparse.COO") if verbose else None
        ## Warn if diagonal is not all ones. It will be converted
        if not np.allclose(np.array(ss[range(ss.shape[0]), range(ss.shape[0])]), 1):
            print("Warning: Similarity matrix diagonal is not all ones. Will set diagonal to all ones.") if verbose else None
        ## Warn if there are any values greater than 1
        if (ss > 1).sum() > 0:
            print("Warning: Similarity matrix has values greater than 1.") if verbose else None
        ## Warn if there are NaNs. Set to 0.
        if (np.isnan(ss.data)).sum() > 0:
            print("Warning: Similarity matrix has NaNs. Will set to 0.") if verbose else None
            ss.data[np.isnan(ss.data)] = 0

    ## Make a boolean matrix for labels
    l_bool = sparse.COO(np.stack([l_arr == u for u in l_u], axis=0))
    samp_per_clust = l_bool.sum(1).todense()
    n_clusters = len(samp_per_clust)
    n_samples = ss.shape[0]
    
    ## Force diagonal to be 1s
    ss = ss.tolil()
    ss[range(n_samples), range(n_samples)] = 1
    ss = sparse.COO(ss)

    ## Compute the similarity matrix for each pair of clusters
    s_big_conj = ss[None,None,:,:] * l_bool[None,:,:,None] * l_bool[:,None,None,:]  ## shape: (n_clusters, n_clusters, n_samples, n_samples)
    s_big_diag = sparse.eye(n_samples) * l_bool[None,:,:,None] * l_bool[:,None,None,:]

    ## Compute the mean similarity matrix for each cluster
    samp_per_clust_crossGrid = samp_per_clust[:,None] * samp_per_clust[None,:]  ## shape: (n_clusters, n_clusters). This is the product of the number of samples in each cluster. Will be used to divide by the sum of similarities.
    norm_mat = samp_per_clust_crossGrid.copy()  ## above variable will be used again and this one will be mutated.
    fixed_diag = samp_per_clust * (samp_per_clust - 1)  ## shape: (n_clusters,). For the diagonal, we need to subtract 1 from the number of samples in each cluster because samples have only 1 similarity with themselves along the diagonal.
    norm_mat[range(n_clusters), range(n_clusters)] = fixed_diag  ## Correcting the diagonal
    s_big_sum_raw = s_big_conj.sum(axis=(2,3)).todense()
    s_big_sum_raw[range(n_clusters), range(n_clusters)] = s_big_sum_raw[range(n_clusters), range(n_clusters)] - samp_per_clust  ## subtract off the number of samples in each cluster from the diagonal
    cs_mean = s_big_sum_raw / norm_mat  ## shape: (n_clusters, n_clusters). Compute mean by finding the sum of the similarities and dividing by the norm_mat.

    ## Compute the min similarity matrix for each cluster
    ### This is done in two steps:
    #### 1. Compute the minimum similarity between each pair of clusters by inverting the similarity matrix and finding the maximum similarity between each pair of clusters.
    #### 2. Since the first step doesn't invert any values that happen to be 0 (since they are sparse), we need to find out if there are any 0 values there are in each cluster pair, and if there then the minimum similarity between the two clusters is 0.
    val_max = s_big_conj.max() + 1
    cs_min = s_big_conj.copy()
    cs_min.data = val_max - cs_min.data  ## Invert the values
    cs_min = cs_min.max(axis=(2,3))  ## Find the max similarity
    cs_min.data = val_max - cs_min.data  ## Invert the values back
    cs_min.fill_value = 0.0  ## Set the fill value to 0.0 since it gets messed up by these subtraction operations
    
    n_missing_values = (samp_per_clust_crossGrid - (s_big_conj > 0).sum(axis=(2,3)).todense())  ## shape: (n_clusters, n_clusters). Compute the number of missing values by subtracting the number of non-zero values from the number of samples in each cluster.
    # n_missing_values[range(len(samp_per_clust)), range(len(samp_per_clust))] = (samp_per_clust**2 - samp_per_clust) - ((s_big_conj[range(len(samp_per_clust)), range(len(samp_per_clust))] > 0).sum(axis=(1,2))).todense()  ## Correct the diagonal by subtracting the number of non-zero values from the number of samples in each cluster. This is because the diagonal is the number of samples in each cluster squared minus the number of samples in each cluster.
    bool_nonMissing_values = (n_missing_values == 0)  ## shape: (n_clusters, n_clusters). Make a boolean matrix for where there are no missing values.
    cs_min = cs_min.todense() * bool_nonMissing_values  ## Set the minimum similarity to 0 where there are missing values.

    ## Compute the max similarity matrix for each cluster
    cs_max = (s_big_conj - s_big_diag).max(axis=(2,3))

    return cs_mean, cs_max.todense(), cs_min


class cDBSCAN():
    """
    Consensus DBSCAN algorithm.
    Runs DBSCAN on each epsilon value and returns the unique
    clusters as well as the number of times each cluster was
    found across epsilon values.
    RH 2022
    
    Args:
        X (np.ndarray):
            The data matrix.
        eps_values (list of float):
            The epsilon values to use in DBSCAN.
        min_samples (int):
            The minimum number of samples to use in DBSCAN.
        **kwargs_dbscan (dict):
            The optional arguments to pass to DBSCAN.
            Includes:
                metric_params (dict): default=None
                    The metric parameters to pass to DBSCAN.
                algorithm (str): default='auto'
                    The algorithm to use in DBSCAN.
                leaf_size (int): default=30
                    The leaf size to use in DBSCAN.
                p (int): default=2
                    The p value to use in DBSCAN.
                n_jobs (int): default=-1
                    The number of jobs to use in DBSCAN.   
    Returns:
        clusters_idx_unique (list of np.ndarray):
            The sample indices for each unique cluster
        clusters_idx_unique_freq (list of int):
            The number of times each unique cluster was found
            over the different epsilon values.
    """
    def __init__(
        self,
        eps_values=[0.1, 0.5, 2.5],
        min_samples=2,
        **kwargs_dbscan
    ):

        # eps_values = np.arange(1.0, 5.0, 0.1)
        # X = block['embeddings']
        # min_samples = 2,

        # kwargs_dbscan = {
        #     'metric_params': None, 
        #     'algorithm': 'auto',
        #     'leaf_size': 30, 
        #     'p': 2, 
        #     'n_jobs': -1
        # }

        self.dbscan_objs = [sklearn.cluster.DBSCAN(
                eps=eps,
                min_samples=min_samples, 
                **kwargs_dbscan
            ) for eps in eps_values]

    def _labels_to_idx(self, labels):
    #     return {label: np.where(labels==label)[0] for label in np.unique(labels)}
        return [np.where(labels==label)[0] for label in np.unique(labels)]

    def _freq_of_values(self, vals):
        u = np.unique(vals)
        f = np.array([np.sum(vals==unique) for unique in u])
        return np.array([f[u==val][0] for val in vals])
            

    def fit(self, X):
        # cDBSCAN: 'consensus DBSCAN': runs a DBSCAN over a sweep of epsilon values and records how many times each cluster appeared
        cluster_idx = []
        for ii, dbscan in enumerate(self.dbscan_objs):
            ## DBSCAN
            db = copy.deepcopy(dbscan)
            db.fit(X) # note that db.labels_==-1 means no cluster found
            db.labels_[self._freq_of_values(db.labels_) < dbscan.min_samples] = -1 # fail safe because sometimes there are clusters of just 1 for some reason...

            # concatenate the indices of clusters found at this eps value to a big list
            [cluster_idx.append(idx) for idx in self._labels_to_idx(db.labels_)]

        clusterHashes_block = [hash(tuple(vec)) for vec in cluster_idx]
        u, idx, c = np.unique(
            ar=clusterHashes_block,
            return_index=True,
            return_counts=True,
        )
        self.clusters_idx_unique = np.array(cluster_idx, dtype=object)[idx]
        self.clusters_idx_unique_freq = c

        return self.clusters_idx_unique, self.clusters_idx_unique_freq


# class Constrained_rich_clustering:
#     """
#     Class to perform constrained cluster assignment.
#     This method takes in putative clusters, a cluster similarity matrix,
#      and a vector of cluster 'scores' describing how valuable each
#      cluster is. It then attempts to find an optimal combination of
#      clusters. This is done by minimizing inclusion of similar clusters
#      and maximizing inclusion of clusters with high scores.
#     The cluster similarity matrix and score vector can be made using 
#      the cluster_dispersion_score and cluster_silhouette_score functions
#      respectively. The cluster membership matrix showing putative clusters
#      can be made by by doing something like sweeping over linkage distances.

#     RH 2022
#     """
#     def __init__(
#         self,
#         c,
#         h,
#         w=None,
#         m_init=None,
#         device='cpu',
#         optimizer_partial=None,
#         scheduler_partial=None,
#         dmCEL_temp=1,
#         dmCEL_sigSlope=2,
#         dmCEL_sigCenter=0.5,
#         dmCEL_penalty=1,
#         sampleWeight_softplusKwargs={'beta': 500, 'threshold': 50},
#         sampleWeight_penalty=1e1,
#         fracWeighted_goalFrac=1.0,
#         fracWeighted_sigSlope=2,
#         fracWeighted_sigCenter=0.5,
#         fracWeight_penalty=1e0,
#         maskL1_penalty=0e-3,
#         tol_convergence=1e-2,
#         window_convergence=100,
#         freqCheck_convergence=100,
#         verbose=True,
#     ):
#         """
#         Initialize the cluster assignment class.

#         Args:
#             c (scipy.sparse.csr_matrix, dtype float):
#                 'Cluster similarity' matrix.
#                 Elements are pairwise similarities between clusters.
#                 shape: (n_clusters, n_clusters)
#             h (scipy.sparse.csr_matrix, dtype bool):
#                 'multi Hot' boolean matrix.
#                 The cluster membership matrix.
#                 shape: (n_samples, n_clusters)
#             w (torch.Tensor, dtype float):
#                 'Weight' vector.
#                 Elements are the scores of each cluster. Can be customized.
#                 Weighs how much to value the inclusion of each cluster
#                  relative to the total fracWeight_penalty.
#                 shape: (n_clusters)
#                 If None, the clusters are weighted equally.
#             m_init (torch.Tensor, dtype float):
#                 'initial Mask' vector.
#                 This vector initializes the only optimized parameter.
#                 The initial cluster inclusion vector. Masks which clusters
#                  are included in the prediction.
#                 shape: (n_clusters)
#                 If None, then vector is initialized as small random values.
#             device (str):
#                 The device to use for the computation. ('cpu', 'cuda', etc.)
#             optimizer_partial (torch.optim.Optimizer):
#                 A torch optimizer with all but the parameters initialized.
#                 If None, then the optimizer is initialized with Adam and 
#                  some default parameters.
#                 This can be made using:
#                  functools.partial(torch.optim.Adam, lr=x, betas=(x,x)).
#             scheduler (torch.optim.lr_scheduler):
#                 A torch learning rate scheduler.
#             dmCEL_temp (float):
#                 ===IMPORTANT HYPERPARAMETER===
#                 The temperature used in the cross entropy loss of the
#                  interaction matrix.
#             dmCEL_sigSlope (float):
#                 The slope of the sigmoid used to constrain and activate the
#                  m vector. Higher values result in more forced certainty, 
#                  but less stability.
#                 Recommend staying as low as possible, around 1-5.
#             dmCEL_sigCenter (float):
#                 The center of the sigmoid described above.
#                 Recommend keeping to 0.5.
#             dmCEL_penalty (float):
#                 The penalty applied to the cross entropy loss of the 
#                  interaction matrix. Best to keep this to 1 and change the
#                  other penalties.
#             sampleWeight_softplusKwargs (dict):
#                 The kwargs passed to the softplus function used to penalize
#                  when samples are included more than once. 
#                 Recommend keeping this around default values.
#             sampleWeight_penalty (float):
#                 ===IMPORTANT HYPERPARAMETER===
#                 The penalty applied when samples are included more than once.
#             fracWeighted_goalFrac (float):
#                 ===IMPORTANT HYPERPARAMETER===
#                 The goal fraction of samples to be included in the output.
#                 Recommend keeping this to 1 for most applications.
#             fracWeighted_sigSlope (float):
#                 The slope of the sigmoid used to constrain the sample weights
#                  so as to give a score simply on whether or not they were
#                  included.
#                 Recommend keeping this to 1-5.
#             fracWeighted_sigCenter (float):
#                 The center of the sigmoid described above.
#                 Recommend keeping to 0.5.
#             fracWeight_penalty (float):
#                 The penalty applied to the sample weights loss.
#             maskL1_penalty (float):
#                 ===IMPORTANT HYPERPARAMETER===
#                 The penalty applied to the L1 norm of the mask.
#                 Adjust this to reduce the probability of extracting clusters
#                  with low scores.
#             tol_convergence (float):
#                 The tolerance for the convergence of the optimization.
#             window_convergence (int):
#                 The number of iterations to use in the convergence check.
#                 A regression is performed on the last window_convergence to
#                  see if the best fit line has changed by less than
#                  tol_convergence.
#             freqCheck_convergence (int):
#                 The number of iterations to wait betwee checking for
#                  convergence.
#             verbose (bool):
#                 Whether to print progress.

#         Attributes set:
#             self.c (torch_sparse.SparseTensor):
#                 The cluster similarity matrix.
#             self.h (torch.SparseTensor):
#                 The cluster membership matrix.
#             self.w (torch.SparseTensor):
#                 The cluster weights vector.
#             self.m (torch.Tensor):
#                 The cluster inclusion vector.
#         """
#         import torch_sparse as ts
#         import scipy.sparse

#         ## Imports
#         super().__init__()

#         ## c.shape[0] == c.shape[1] == h.shape[1]
#         assert c.shape[0] == c.shape[1] == h.shape[1], 'RH ERROR: the following must be true:  c.shape[0] == c.shape[1] == h.shape[1], they should all be n_clusters'

#         self._n_clusters = h.shape[1]

#         self._DEVICE = device

#         ## c must be cast as a torch_sparse.SparseTensor because the masking 
#         ##  operation (c * m[None,:]) is not implemented in standard torch.sparse yet...
#         c_tmp = indexing.scipy_sparse_to_torch_coo(c).coalesce().type(torch.float32)
#         self.c = ts.SparseTensor(
#             row=c_tmp.indices()[0], 
#             col=c_tmp.indices()[1], 
#             value=c_tmp.values(),
#             sparse_sizes=c_tmp.shape,
#         ).to(self._DEVICE)

#         self.h = indexing.scipy_sparse_to_torch_coo(h).coalesce().type(torch.float32).to(self._DEVICE)

#         ## w is converted to a square diagonal matrix so that h @ w masks the columns of h (torch.sparse doesn't support h * w[None,:] yet...)
#         w_tmp = scipy.sparse.lil_matrix((self._n_clusters, self._n_clusters))
#         w_tmp[range(self._n_clusters), range(self._n_clusters)] = (w / w.max()) if w is not None else 1
#         self.w = indexing.scipy_sparse_to_torch_coo(w_tmp).coalesce().type(torch.float32).to(self._DEVICE)

#         ## optimization seems to proceed best when values initialize as small random values. New ideas welcome here.
#         self.m = m_init.to(self._DEVICE) if m_init is not None else (torch.ones(self._n_clusters)*0.1 + torch.rand(self._n_clusters)*0.05).type(torch.float32).to(self._DEVICE)
#         self.m.requires_grad=True

#         self._dmCEL_penalty = dmCEL_penalty
#         self._sampleWeight_penalty = sampleWeight_penalty
#         self._fracWeight_penalty = fracWeight_penalty
#         self._maskL1_penalty = maskL1_penalty

#         self._dmCEL_temp = torch.as_tensor([dmCEL_temp], dtype=torch.float32).to(self._DEVICE)
#         self._dmCEL_sigSlope = torch.as_tensor([dmCEL_sigSlope], dtype=torch.float32).to(self._DEVICE)
#         self._dmCEL_sigCenter = torch.as_tensor([dmCEL_sigCenter], dtype=torch.float32).to(self._DEVICE)
#         self._fracWeighted_sigSlope = torch.as_tensor([fracWeighted_sigSlope], dtype=torch.float32).to(self._DEVICE)
#         self._fracWeighted_sigCenter = torch.as_tensor([fracWeighted_sigCenter], dtype=torch.float32).to(self._DEVICE)


#         # self._dmCEL_penalty  = torch.as_tensor([self._dmCEL_penalty]).to(self._DEVICE)
#         # self._sampleWeight_penalty = torch.as_tensor([self._sampleWeight_penalty]).to(self._DEVICE)
#         # self._fracWeight_penalty = torch.as_tensor([self._fracWeight_penalty]).to(self._DEVICE)
#         # self._maskL1_penalty = torch.as_tensor([self._maskL1_penalty]).to(self._DEVICE)

#         # self._dmCEL_penalty.requires_grad = True
#         # self._sampleWeight_penalty.requires_grad = True
#         # self._fracWeight_penalty.requires_grad = True
#         # self._maskL1_penalty.requires_grad = True

#         # self._dmCEL_temp.requires_grad = True
#         # self._dmCEL_sigSlope.requires_grad = True
#         # self._dmCEL_sigCenter.requires_grad = True
#         # self._fracWeighted_sigSlope.requires_grad = True
#         # self._fracWeighted_sigCenter.requires_grad = True

#         self._optimizer = optimizer_partial(params=[self.m]) if optimizer_partial is not None else torch.optim.Adam(params=[self.m], lr=1e-2, betas=(0.9, 0.900))
#         # self._optimizer = optimizer_partial(params=[self.m, self._dmCEL_temp, self._dmCEL_sigSlope, self._dmCEL_sigCenter, self._fracWeighted_sigSlope, self._fracWeighted_sigCenter]) if optimizer_partial is not None else torch.optim.Adam(params=[self.m], lr=1e-2, betas=(0.9, 0.900))
#         self._scheduler = scheduler_partial(optimizer=self._optimizer) if scheduler_partial is not None else torch.optim.lr_scheduler.LambdaLR(optimizer=self._optimizer, lr_lambda=lambda x : x, last_epoch=-1, verbose=True)
        
#         self._dmCEL = self._DoubleMasked_CEL(
#             c=self.c,
#             device=self._DEVICE,
#             temp=self._dmCEL_temp,
#             sig_slope=self._dmCEL_sigSlope,
#             sig_center=self._dmCEL_sigCenter,
#         )

#         self._loss_fracWeighted = self._Loss_fracWeighted(
#             h=self.h,
#             w=self.w,
#             goal_frac=fracWeighted_goalFrac,
#             sig_slope=self._fracWeighted_sigSlope,
#             sig_center=self._fracWeighted_sigCenter,
#         )

#         self._loss_sampleWeight = self._Loss_sampleWeight(
#             softplus_kwargs=sampleWeight_softplusKwargs,
#         )

#         self._convergence_checker = self._Convergence_checker(
#             tol_convergence=tol_convergence,
#             window_convergence=window_convergence,
#         )

#         self._tol_convergence = tol_convergence
#         self._window_convergence = window_convergence
#         self._freqCheck_convergence = freqCheck_convergence
#         self._verbose = verbose

#         self._i_iter = 0
#         self.losses_logger = {
#             'loss': [],
#             'L_cs': [],
#             'L_fracWeighted': [],
#             'L_sampleWeight': [],
#             'L_maskL1': [],
#         }

#         ## clean up. Some of the initialization steps generate intermediate tensors
#         ##  that are not needed long term.
#         gc.collect()
#         torch.cuda.empty_cache()

#     def fit(
#         self, 
#         min_iter=1e3,
#         max_iter=1e4,
#         verbose=True, 
#         verbose_interval=100
#     ):
#         """
#         Fit the model.
#         This method can be interupted during a run and restarted without issues.

#         Uses the standard PyTorch gradient descent approach:
#             1. zero the parameter gradients (optimizer.zero_grad())
#             2. compute the loss (loss=...)
#             3. compute the gradients of the loss w.r.t. the parameters (loss.backward())
#             4. update the parameters (optimizer.step())

#         There are 4 components to the loss:
#             1. L_cs: Penalizes how similar the unmasked clusters are relative to each other.
#             2. L_fracWeighted: Penalizes how many samples are not assigned to a cluster.
#             3. L_sampleWeight: Penalizes samples that are assigned to multiple clusters.
#             4. L_maskL1: Penalizes using many (instead of few) clusters to mask the samples.

#         Issue handling things:
#             - clamping the values of m to > -14. This is to prevent a dead gradient on those
#              parameter values.
#             - checking if loss is NaN. If it is, the optimization is terminated before backward pass.

#         The convergence checker fits a line to the last self._window_convergence iterations of self.losses_logger
#          and checks if the slope of the line is less than self._tol_convergence.

#         Args:
#             min_iter (int): 
#                 The minimum number of iterations to run.
#             max_iter (int): 
#                 The maximum number of iterations to run.
#             verbose (bool): 
#                 If True, print the loss every verbose_interval iterations.
#             verbose_interval (int): 
#                 The interval at which to print the loss.
#         """
#         loss_smooth = np.nan
#         diff_window_convergence = np.nan
#         while self._i_iter <= max_iter:
#             self._optimizer.zero_grad()

#             L_cs = self._dmCEL(c=self.c, m=self.m, w=self.w) * self._dmCEL_penalty  ## 'cluster similarity loss'
#             L_sampleWeight = self._loss_sampleWeight(self.h, self.activate_m()) * self._sampleWeight_penalty ## 'sample weight loss'
#             L_fracWeighted = self._loss_fracWeighted(self.activate_m()) * self._fracWeight_penalty ## 'fraction weighted loss'
#             L_maskL1 = torch.sum(torch.abs(self.activate_m())) * self._maskL1_penalty ## 'L1 on mask loss'

#             self._loss = L_cs + L_fracWeighted + L_sampleWeight + L_maskL1 ## 'total loss'

#             if torch.isnan(self._loss):
#                 print(f'STOPPING EARLY: loss is NaN. iter: {self._i_iter}  loss: {self._loss.item():.4f}  L_cs: {L_cs.item():.4f}  L_fracWeighted: {L_fracWeighted.item():.4f}  L_sampleWeight: {L_sampleWeight.item():.4f}  L_maskL1: {L_maskL1.item():.4f}')
#                 break

#             self._loss.backward()
#             self._optimizer.step()
#             self._scheduler.step()

#             self.m.data = torch.maximum(self.m.data , torch.as_tensor(-14, device=self._DEVICE)) ## clamp m values to prevent zeroing out and a dead gradient

#             ## populate logger with values
#             self.losses_logger['loss'].append(self._loss.item())
#             self.losses_logger['L_cs'].append(L_cs.item())
#             self.losses_logger['L_fracWeighted'].append(L_fracWeighted.item())
#             self.losses_logger['L_sampleWeight'].append(L_sampleWeight.item())
#             self.losses_logger['L_maskL1'].append(L_maskL1.item())

#             ## check for convergence
#             if self._i_iter%self._freqCheck_convergence==0 and self._i_iter>self._window_convergence and self._i_iter>min_iter:
#                 diff_window_convergence, loss_smooth, converged = self._convergence_checker(self.losses_logger['loss'])
#                 if converged:
#                     print(f"STOPPING: Convergence reached in {self._i_iter} iterations.  loss: {self.losses_logger['loss'][-1]:.4f}  loss_smooth: {loss_smooth:.4f}")
#                     break
            
#             ## print loss values
#             if verbose and self._i_iter % verbose_interval == 0:
#                 print(f'iter: {self._i_iter}:  loss_total: {self._loss.item():.4f}  lr: {self._scheduler.get_last_lr()[0]:.5f}   loss_cs: {L_cs.item():.4f}  loss_fracWeighted: {L_fracWeighted.item():.4f}  loss_sampleWeight: {L_sampleWeight.item():.4f}  loss_maskL1: {L_maskL1.item():.4f}  diff_loss: {diff_window_convergence:.4f}  loss_smooth: {loss_smooth:.4f}')
#             self._i_iter += 1


#     def predict(
#         self,
#         m_threshold=0.5
#     ):   
#         """
#         Return predicted cluster assignments based on a threshold on the
#          activated 'm' vector.
#         It can be useful to first plot the activated 'm' vector to see the
#          distribution: self.plot_clusterWeights().
        
#         Args:
#             m_threshold (float): 
#                 Threshold on the activated 'm' vector.
#                 Clusters with activated 'm' values above the threshold are
#                  assigned a cluster ID.
#         """
#         h_ts = indexing.torch_to_torchSparse(self.h)

#         self.m_bool = (self.activate_m() > m_threshold).squeeze().cpu()
#         h_ts_bool = h_ts[:, self.m_bool]
#         n_multiples = (h_ts_bool.sum(1) > 1).sum()
#         print(f'WARNING: {n_multiples} samples are matched with multiple clusters. Consider increasing the sample_weight_penalty during training.') if n_multiples > 0 else None
#         h_preds = (h_ts_bool * (torch.arange(self.m_bool.sum(), device=self._DEVICE)[None,:]+1)).detach().cpu()

#         if h_preds.numel() == 0:
#             print(f'WARNING: No predictions made.  m_threshold: {m_threshold}')
#             return None, None

#         preds = h_preds.max(dim=1) - 1
#         preds[torch.isinf(preds)] = -1
        
#         h_m = (h_ts * self.activate_m()[None,:]).detach().cpu().to_dense()
#         confidence = h_m.var(1) / h_m.mean(1)

#         self.scores_clusters = torch_helpers.diag_sparse(self.w).cpu()[self.m_bool]
#         self.scores_samples = (h_preds * self.scores_clusters[None,:]).max(1)
        
#         self.preds = preds
#         self.confidence = confidence
        
#         return self.preds, self.confidence, self.scores_samples, self.m_bool

#     def activate_m(self):
#         """
#         Pass the 'm' ('masking', the only optimized parameter) through the
#          activation function and return it.
#         """
#         return self._dmCEL.activation(self.m)


#     class _DoubleMasked_CEL:
#         """
#         'Double Masked Cross-Entropy Loss' Class. Derived here.
#         Gives a loss value for how well-separated a set of clusters are.
#         Input is a cluster similarity matrix (c, shape: (n_clusters, n_clusters)),
#          and a mask vector (m, shape: (n_clusters,)).
#         Output is a loss value.
#         Operation proceeds as follows:
#         1. Pass mask vector through activation function. ma=sigmoid(m)
#         2. Compute the loss value for each cluster.
#             a. Mask the columns of 'c' with 'ma': cm=c*ma[None,:]
#             b. Replace the diagonal values of 'cm' with the diagonal values of 'c'
#             c. Compute the cross-entropy loss value for each cluster. lv=CEL(cm, labels=arange(n_clusters))
#         3. Weight the loss values by the mask vector and do a mean reduction. l=lv@ma
        
#         RH 2022
#         """
#         def __init__(
#             self,
#             c,
#             device='cpu',
#             temp=1,
#             sig_slope=5,
#             sig_center=0.5,
#         ):    
#             """
#             Initializes the 'Double Masked Cross-Entropy Loss' Class.

#             Args:
#                 c (torch_sparse.SparseTensor):
#                     Cluster similarity matrix (c, shape: (n_clusters, n_clusters)).
#                 device (str):
#                     Device to use for the computation.
#                 temp (float):
#                     Temperature parameter for the softmax activation function.
#                     Lower values results in more non-convexity but stronger convergence (steeper gradient).
#                 sig_slope (float):
#                     Slope of the sigmoid activation function of the mask vector.
#                 sig_center (float):
#                     Center of the sigmoid activation function of the mask vector.
#             """
#             self._n_clusters = c.sizes()[0]

#             self.temp = temp
            
#             ## Cross-Entropy Loss: since we are trying to maximize the diagonal, CEL can be simplified as:
#             ##  -diag(log_softmax(x))
#             self.CEL = lambda x :  - ts_logSoftmax(x, temperature=self.temp, shift=None).get_diag() ## this function ts_logSoftmax existed at one point
            
#             ## We will be activating the optimized 'm' parameter to scale it between 0 and 1.
#             self.activation = self.make_sigmoid_function(sig_slope, sig_center)

#             self.device=device
            
#             ## We define a best and worst case CEL value so that we can scale the loss to be invariant to
#             ##  things like temperature and number of clusters.
#             self.worst_case_loss = self.CEL(c)
#             self.best_case_loss = self.CEL(
#                 ts.tensor.SparseTensor(
#                     row=torch.arange(self._n_clusters, device=device, dtype=torch.int64),
#                     col=torch.arange(self._n_clusters, device=device, dtype=torch.int64),
#                     value=c.get_diag().type(torch.float32)**0,
#                     sparse_sizes=(self._n_clusters, self._n_clusters),
#                 )
#             ) ## the best case is just a diagonal matrix with the same values as the input matrix.
#             self.worst_minus_best = self.worst_case_loss - self.best_case_loss  + 1e-8  ## add a small amount for numerical stability.

#         def make_sigmoid_function(
#             self,
#             sig_slope=5,
#             sig_center=0.5,
#         ):
#             return lambda x : 1 / (1 + torch.exp(-sig_slope*(x-sig_center)))
            
#         def __call__(self, c, m, w):
#             import torch_sparse as ts
#             import scipy.sparse

#             ma = self.activation(m)  ## 'activated mask'. Constrained to be 0-1

#             ## Below gives us our loss for each column, which will later be scaled by 'ma' 
#             ##  to get the final loss.
#             ## The idea is that 'ma' masks out columns in 'c' (while ignoring elements along
#             ##  the diagonal) in order to maximize the diagonal elements relative to all other
#             ##  elements in the row. Optimization should result in 'selection' of clusters by
#             ##  masking out poor clusters that are too similar to good clusters.
#             lv = self.CEL(
#                 ts_setDiag_lowMem(
#                     c * ma[None,:],
#                     c.get_diag()**0
#                 )  ## this function existed at one point too
#             ) ## 'loss vector'

#             lv_norm = (lv - self.best_case_loss) / self.worst_minus_best  ## scale the loss of each cluster by the best and worst case possibilities.
            
#             ## weight the loss of each clusters by the activated masking parameter 'ma'
#             l = (lv_norm @ ma) / ma.sum()  ## 'loss'

#             return l


#     class _Loss_fracWeighted:
#         """
#         'Fraction Weighted Loss' Class. Derived here.
#         Gives a loss value for the fraction of clusters that are highly weighted
#          scaled by how valuable each cluster is.
#         Inputs are the cluster membership ('multiHot') matrix (h, shape: (n_clusters, n_samples)),
#          and the cluster scores ('weighting') vector (w, shape: (n_clusters,)).
#         Output is a loss value.
#         Operation proceeds as follows:
#         0. In initialization, weight the h matrix by the w vector. hw=h*w[None,:]
#         1. Mask and sum-reduce the hw matrix with the mask vector (dot product) to get the
#          score-weighted penalty for excluding that cluster. hwm=hw@m (hwm shape: n_samples)
#         2. Pass hwm through an activation function. hwma=sigmoid(hwm)
#         3. Take mean squared error with respect to the goal number of samples positively
#          weighted. l=(hwma-g)^2

#         RH 2022
#         """
#         def __init__(
#             self,
#             h,
#             w,
#             goal_frac=1.0,
#             sig_slope=5,
#             sig_center=0.5,
#         ):
#             # self.h_w = h.type(torch.float32) * w[None,:]
#             self.h_w = h.type(torch.float32) @ w
#             # self.h_w = h.type(torch.float32)
        
#             self.goal_frac = goal_frac
#             self.sigmoid = self.make_sigmoid_function(sig_slope, sig_center)
            
#         def make_sigmoid_function(
#             self,
#             sig_slope=5,
#             sig_center=0.5,
#         ):
#             return lambda x : 1 / (1 + torch.exp(-sig_slope*(x-sig_center)))

#         def generate_sampleWeights(self, m):
#             return self.h_w @ m
#         def activate_sampleWeights(self, sampleWeights):
#             return self.sigmoid(sampleWeights)

#         def __call__(self, m):
#             return ((self.activate_sampleWeights(self.generate_sampleWeights(m))).mean() - self.goal_frac)**2

#     class _Loss_sampleWeight:
#         def __init__(
#             self,
#             softplus_kwargs=None,
#         ):
#             self.softplus = torch.nn.Softplus(**softplus_kwargs)
#             self.weight_high = 1
#             self.offset = self.activate_sampleWeights(torch.ones(1, dtype=torch.float32)).max()
        
#         def generate_sampleWeights(self, h, m):
#             return h @ m
#         def activate_sampleWeights(self, sampleWeights):
#             return self.softplus(sampleWeights - 1 - 4/self.softplus.beta)*self.weight_high

#         def __call__(self, h, m):
#             return (self.activate_sampleWeights(self.generate_sampleWeights(h, m))).mean()

#     class _Convergence_checker:
#         """
#         'Convergence Checker' Class.
#         Checks for convergence of the optimization. Uses Ordinary Least Squares (OLS) to 
#          fit a line to the last 'window_convergence' number of iterations.
#         """
#         def __init__(
#             self,
#             tol_convergence=1e-2,
#             window_convergence=100,
#         ):
#             """
#             Initialize the convergence checker.
            
#             Args:
#                 tol_convergence (float): 
#                     Tolerance for convergence.
#                     Corresponds to the slope of the line that is fit.
#                 window_convergence (int):
#                     Number of iterations to use for fitting the line.
#             """
#             self.window_convergence = window_convergence
#             self.tol_convergence = tol_convergence

#             self.line_regressor = torch.cat((torch.linspace(0,1,window_convergence)[:,None], torch.ones((window_convergence,1))), dim=1)

#         def OLS(self, y):
#             """
#             Ordinary least squares.
#             Fits a line to y.
#             """
#             X = self.line_regressor
#             theta = torch.inverse(X.T @ X) @ X.T @ y
#             y_rec = X @ theta
#             bias = theta[-1]
#             theta = theta[:-1]

#             return theta, y_rec, bias

#         def __call__(
#             self,
#             loss_history,
#         ):
#             """
#             Forward pass of the convergence checker.
#             Checks if the last 'window_convergence' number of iterations are
#              within 'tol_convergence' of the line fit.

#             Args:
#                 loss_history (list):
#                     List of loss values for the last 'window_convergence' number of iterations.

#             Returns:
#                 diff_window_convergence (float):
#                     Difference of the fit line over the range of 'window_convergence'.
#                 loss_smooth (float):
#                     The mean loss over 'window_convergence'.
#                 converged (bool):
#                     True if the 'diff_window_convergence' is less than 'tol_convergence'.
#             """
#             if len(loss_history) < self.window_convergence:
#                 return torch.nan, torch.nan, False
#             loss_window = torch.as_tensor(loss_history[-self.window_convergence:], device='cpu', dtype=torch.float32)
#             theta, y_rec, bias = self.OLS(y=loss_window)

#             diff_window_convergence = (y_rec[-1] - y_rec[0])
#             loss_smooth = loss_window.mean()
#             converged = True if torch.abs(diff_window_convergence) < self.tol_convergence else False
#             return diff_window_convergence.item(), loss_smooth.item(), converged
            

#     def plot_loss(self):
#         plt.figure()
#         plt.plot(self.losses_logger['loss'], linewidth=4)
#         plt.plot(self.losses_logger['L_cs'])
#         plt.plot(self.losses_logger['L_fracWeighted'])
#         plt.plot(self.losses_logger['L_sampleWeight'])
#         plt.plot(self.losses_logger['L_maskL1'])

#         plt.legend(self.losses_logger.keys())
#         plt.xlabel('iteration')
#         plt.ylabel('loss')

#     def plot_clusterWeights(self, plot_raw_m=False):
#         plt.figure()
#         if plot_raw_m:
#             plt.hist(self.m.detach().cpu().numpy(), bins=50)
#         plt.hist(self.activate_m().detach().cpu().numpy(), bins=50)
#         plt.xlabel('cluster weight')
#         plt.ylabel('count')

#     def plot_clusterScores(self, bins=100):
#         if hasattr(self, 'm_bool'):
#             m_bool = self.m_bool
#             confidence = self.confidence
#             preds = self.preds
#         else:
#             preds, confidence, scores_samples, m_bool = self.predict()
#             if preds is None:
#                 print('Plot failed: preds is None.')
#                 return None
#         scores = helpers.diag_sparse(self.w).cpu()[m_bool.cpu()]

#         plt.figure()
#         plt.hist(scores, bins=bins, log=True)
#         plt.xlabel('cluster score')
#         plt.ylabel('count')

#         plt.figure()
#         plt.hist(confidence[preds >= 0], bins=bins, log=True)
#         plt.xlabel('confidence')
#         plt.ylabel('count')

#     def plot_sampleWeights(self):
#         sampleWeights = self._loss_sampleWeight.generate_sampleWeights(self.h, self.activate_m()).detach().cpu().numpy()

#         plt.figure()
#         plt.hist(sampleWeights, bins=50)
#         plt.xlabel('sample weight')
#         plt.ylabel('count')

#     def plot_labelCounts(self):
#         if hasattr(self, 'preds'):
#             preds = self.preds
#         else:
#             preds, confidence, scores_samples, m_bool = self.predict()
#             if preds is None:
#                 print('Skipping plot_labelCounts: preds is None.')
#                 return None
            
#         labels_unique, label_counts = np.unique(preds, return_counts=True)

#         fig, axs = plt.subplots(1, 2, figsize=(10,5))
#         axs[0].bar(labels_unique, label_counts)
#         axs[0].set_xlabel('label')
#         axs[0].set_ylabel('count')
#         axs[1].hist(label_counts[labels_unique>=0], bins=25)
#         axs[1].set_xlabel('n_roi per cluster')
#         axs[1].set_ylabel('counts')

#         return fig, axs

#     def plot_c_threshold_matrix(self, m_threshold=0.5):
#         mt = self.activate_m() > m_threshold
#         plt.figure()
#         plt.imshow(self.c[mt, mt].detach().cpu(), aspect='auto')

#     def plot_c_masked_matrix(self, m_threshold=0.5, **kwargs_imshow):
#         import sparse
#         mt = self.activate_m() > m_threshold
#         # return (self.c * mt[None,:])
#         cdm = (sparse.COO(self.c.to_scipy()) * mt[None,:].cpu().numpy()  * mt[:, None].cpu().numpy()).tocsr().tolil()
#         cdm[list(self._dmCEL.idx_diag.cpu().numpy()), list(self._dmCEL.idx_diag.cpu().numpy())] = self.c.get_diag().cpu()
#         plt.figure()
#         plt.imshow(cdm.toarray(), aspect='auto')


# class rich_clust():
#     """
#     Basic clustering algorithm based on gradient descent.
#     The approach is to optimize a 'multihot matrix' (h),
#      which is an n_samples x n_clusters matrix describing
#      the membership of each sample to each cluster. This matrix
#      is used to mask a similarity matrix (s); allowing for the
#      calculation of the mean pairwise similarity within each 
#      cluster and between each pair of clusters. The loss
#      function maximizes the within similarity and minimizes the
#      between similarity.
#     Functionally, the result ends up similar to k-means
#      clustering.
#     This method is not great on its own, but is useful because
#      it allows for custom penalties to be applied to the loss.
#     It struggles due to a non-convex solution space which
#      results in sensitivity to initial conditions and weird 
#      sensitivities to hyperparameters.
#     RH 2022
#     """

#     def __init__(
#         self,
#         s,
#         n_clusters=2,
#         l=1,
#         temp_h=1,
#         temp_c=1,
#         optimizer=None,
#         DEVICE='cpu',
#         init_h=None,
#     ):
#         """
#         Args:
#             s (torch.Tensor):
#                 Similarity matrix.
#                 shape: (n_samples, n_samples)
#                 dtype: torch.float32
#                 The diagonal should be zeros.
#                 Example:
#                     d = torch.cdist(data, data, p=2).type(torch.float32)
#                     s = torch.maximum(1-d, torch.as_tensor([0]))
#                     s = s * torch.logical_not(torch.eye(s.shape[0]))
#             n_clusters (int):
#                 The number of clusters to find.
#             l (float):
#                 'Locality' parameter. The exponent applied to 
#                   the similarity matrix. Higher values allow for
#                   more non-convex clusters, but can result in 
#                   instability and cluster-splitting.
#             temp_h (float):
#                 The temperature for the multihot matrix. Higher
#                  values result in fuzzier cluster edges.
#             temp_c (float):
#                 The temperature for the cluster membership matrix.
#                 Higher values result in globally less confident
#                  cluster scores.
#             optimizer (partial torch.optim.Optimizer):
#                 Optional. If None, then Adam is used.
#                 The partial optimizer to use.
#                 Can be constructed like: 
#                     functools.partial(torch.optim.Adam, lr=0.01, weight_decay=0.00001)
#             DEVICE (str):
#                 The device to use. Default is 'cpu'.
#             init_h (torch.Tensor):
#                 The initial multihot matrix. If None, then
#                  random initialization is used.
#         """
    
#         self.s = s.to(DEVICE)**l
#         self.n_clusters = n_clusters

#         self.l = l
#         self.temp_h = temp_h
#         self.temp_c = temp_c

#         self.DEVICE = DEVICE

#         if init_h is None:
#             self.h = self._initialize_multihot_matrix(self.s.shape[0], self.n_clusters).to(self.DEVICE)
#         else:
#             self.h = init_h.to(self.DEVICE)
#         self.h.requires_grad = True

#         self.ii_normFactor = lambda i   : i * (i-1)
#         self.ij_normFactor = lambda i,j : i * j

#         if optimizer is None:
#             self.optimizer = torch.optim.Adam(
#                 [self.h], 
#                 lr=0.1,
#                 # weight_decay=1*10**-6
#             )
#         else:
#             self.optimizer = optimizer(params=[self.h])


#     def _initialize_multihot_matrix(self, n_samples, n_clusters):
#         h = torch.rand(size=(n_samples, n_clusters))  ## Random initialization
#         return h
    
#     def _make_cluster_similarity_matrix(self, s, h, temp_h, DEVICE='cpu'):
#         h = torch.nn.functional.softmax(h/temp_h, dim=1)
#     #     return torch.einsum('ab, abcd -> cd', s**1, torch.einsum('ac, bd -> abcd', h,h))  /  ( (torch.eye(h.shape[1]).to(DEVICE) * ii_normFactor(h.sum(0))) + ((1-torch.eye(h.shape[1]).to(DEVICE)) * ij_normFactor(*torch.meshgrid((h.sum(0), h.sum(0)), indexing='ij'))) )
#     #     return (  torch.einsum('ab, abcd -> cd', s**1, torch.einsum('ac, bd -> abcd', h,h)) * torch.eye(h.shape[1]).to(DEVICE)  +  torch.einsum('ab, abcd -> cd', s**4, torch.einsum('ac, bd -> abcd', h,h)) * (torch.logical_not(torch.eye(h.shape[1]).to(DEVICE)))*0.05  )  /  ( (torch.eye(h.shape[1]).to(DEVICE) * ii_normFactor(h.sum(0))) + ((1-torch.eye(h.shape[1]).to(DEVICE)) * ij_normFactor(*torch.meshgrid((h.sum(0), h.sum(0)), indexing='ij'))) )
#         return torch.einsum('ab, ac, bd -> cd', s, h, h)  /  \
#             ( (torch.eye(h.shape[1]).to(DEVICE) * self.ii_normFactor(h.sum(0))) + ((1-torch.eye(h.shape[1]).to(DEVICE)) * self.ij_normFactor(*torch.meshgrid((h.sum(0), h.sum(0)), indexing='ij'))) )

#     def fit(self, n_iter=200):

#         for i_iter in range(n_iter):
#             self.optimizer.zero_grad()

#             self.c = self._make_cluster_similarity_matrix(
#                 s=self.s,
#                 h=self.h, 
#                 temp_h=self.temp_h,
#                 DEVICE=self.DEVICE
#             )
            
#             self.L_cs = torch.nn.functional.cross_entropy(self.c/self.temp_c, torch.arange(self.n_clusters).to(self.DEVICE))
#             self.loss = self.L_cs
#             self.loss.backward()
#             self.optimizer.step()



from __future__ import annotations

from pathlib import Path
import os
import json
import copy
import time
import typing
import sys
import datetime
import getpass

def _require_natsort():
    try:
        import natsort
    except ImportError as e:
        raise ImportError(
            "natsort is required for natural sorting in server helpers. "
            "Install with `pip install bnpm[io]`."
        ) from e
    return natsort


def _require_paramiko():
    try:
        import paramiko
    except ImportError as e:
        raise ImportError(
            "paramiko is required for SSH/SFTP operations. "
            "Install with `pip install bnpm[server]`."
        ) from e
    return paramiko

def batch_run(
    paths_scripts, 
    params_list, 
    sbatch_config_list, 
    max_n_jobs=2,
    dir_save='/n/data1/hms/neurobio/sabatini/rich/analysis/', 
    name_save='jobNum_', 
    verbose=True,
):
    r"""
    Run a batch of jobs.
    Workflow 1: run a single script over a sweep of parameters
        - Make a script that takes in the set of parameters
           you wish to sweep over as variables.
        - Prepend the script to take in string arguments
           pointing to a param_config file (maybe a dict).
           See paths_scripts Arg below for details.
        - Save the script in .py file.
        - In a new script, call this function (batch_run)
        - A new job will be run for each item in params_list
            - Each job will make a new directory, and within
               it will save (1) a .json file containing the 
               parameters used, and (2) the .sh file that 
               was run.
        - Save output files using the 'dir_save' argument

    Alternative workflows where you have multiple different
     scripts or different config files are also possible.

    RH 2021

    Args:
        paths_scripts (List):
            - List of script paths to run.
            - List can contain either 1 or n_jobs items.
            - Each script must save its results it's own way
               using a relative path (see 'dir_save' below)
            - Each script should contain the following to handle
               input arguments specified by the user and this
               function, DEMO:
                ```
                import sys
                    path_script, path_params, dir_save = sys.argv
                
                import json
                with open(path_params, 'r') as f:
                    params = json.load(f)
                ```                
            - It's also good practice to save the script .py file
               within dir_save DEMO:
                ```
                import shutil
                shutil.copy2(
                    path_script, 
                    str(Path(dir_save) / Path(path_script).name)
                    );
                ```
        params_list (List):
            - Parameters (arguments) to be used
            - List can contain either 1 or n_jobs items.
            - Each will be saved as a .json file (so nothing too big)   
            - Will be save into each inner/job directory and the path
               will be passed to the script for each job.
        sbatch_config_list (List):
            - List of string blocks containing the arguments to 
               pass for each job/script.
            - List can contain either 1 or n_jobs items.
            - Must contain: python "$@" at the bottom (to take in 
               arguments), and raw string must have '\n' to signify
               line breaks.
               Demo: '#!/usr/bin/bash
                    #SBATCH --job-name=python_01
                    #SBATCH --output=jupyter_logs/python_01_%j.log
                    #SBATCH --partition=priority
                    #SBATCH -c 1
                    #SBATCH -n 1
                    #SBATCH --mem=1GB
                    #SBATCH --time=0-00:00:10

                    unset XDG_RUNTIME_DIR

                    cd /n/data1/hms/neurobio/sabatini/rich/

                    date

                    echo "loading modules"
                    module load gcc/9.2.0 cuda/11.2

                    echo "activating environment"
                    source activate ROI_env

                    echo "starting job"
                    python "$@" '
        max_n_jobs (int):
            - Maximum number of jobs that can be called
            - Used as a safety precaution
            - Be careful that params_list has the right len
        dir_save (str or Path):
            - Outer directory to save results to.
            - Will be created if it does not exist.
            - Will be populated by folders for each job
            - Will be sent to the script for each job as the
               third argument. See paths_scripts demo for details.
        name_save (str or List):
            - Name of each job (used as inner directory name)
            - If str, then will be used for all jobs 
            - Job iteration always appended to the end.
            - If List, then must have len(params_list) items.
        verbose (bool):
            - Whether or not to print progress
    """
    import json
    import os
    import shutil
    from . import container_helpers

    # make sure the arguments are matched in length
    n_jobs = max(len(paths_scripts), len(params_list), len(sbatch_config_list))
    if max_n_jobs is not None:
        if n_jobs > max_n_jobs:
            raise ValueError(f'Too many jobs requested: max_n_jobs={max_n_jobs} > n_jobs={n_jobs}')

    def rep_inputs(item, n_jobs):
        if len(item)==1 and (n_jobs>1):
            return container_helpers.lazy_repeat_obj(item[0], pseudo_length=n_jobs)
        else:
            return item

    paths_scripts      = rep_inputs(paths_scripts,   n_jobs)
    params_list        = rep_inputs(params_list,  n_jobs)
    sbatch_config_list = rep_inputs(sbatch_config_list, n_jobs)
    name_save          = rep_inputs([name_save], n_jobs)

    # setup the save path
    Path(dir_save).mkdir(parents=True, exist_ok=True)
    dir_save = Path(dir_save).resolve()

    # run the jobs
    for ii in range(n_jobs):
        dir_save_job = dir_save / f'{name_save[ii]}{ii}'
        dir_save_job.mkdir(parents=True, exist_ok=True)
        # save the shell scripts
        save_path_sbatchConfig = dir_save_job / 'sbatch_config.sh'
        with open(save_path_sbatchConfig, 'w') as f:
            f.write(sbatch_config_list[ii])
        # save the script
        path_script_job = dir_save_job / Path(paths_scripts[ii]).name
        shutil.copyfile(paths_scripts[ii], path_script_job);
        # save the parameters        
        path_params_job = dir_save_job / 'params.json'
        with open(path_params_job, 'w') as f:
            json.dump(params_list[ii], f)
    
        # run the job
        if verbose:
            print(f'Submitting job: {name_save[ii]} {ii}')
        # ! sbatch --job-name=${name_save}_${ii} --output=${dir_save_job}/log.txt --error=${dir_save_job}/err.txt --time=${sbatch_config_list[ii]["time"]} --mem=${sbatch_config_list[ii]["mem"]} --cpus-per-task=${sbatch_config_list[ii]["cpus"]} --wrap="${paths_scripts[ii]} ${params_list[ii]} ${sbatch_config_list[ii]} ${dir_save_job}"
        os.system(f'sbatch {save_path_sbatchConfig} {paths_scripts[ii]} {path_params_job} {dir_save_job}')


###############################
### MASTER CONTROLLER STUFF ###
###############################

import time
from pathlib import Path
import os
import stat
import re

class ssh_interface():
    """
    Interface to ssh to a remote server.
    Mostly a wrapper for paramiko.SSHClient.
    Tested on O2 cluster at Harvard.
    RH 2022
    """
    def __init__(
        self,
        nbytes_toReceive=4096,
        recv_timeout=1,
        verbose=True,
    ):
        """
        Args:
            nbytes_toReceive (int):
                Number of bytes to receive at a time.
                Caps the maximum message it can receive.
            recv_timeout (int):
                Timeout for receiving data.
            verbose (bool):
                Whether or not to print progress
        """
        
        self.nbytes = nbytes_toReceive
        self.recv_timeout = recv_timeout
        self.verbose=verbose

        paramiko = _require_paramiko()
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    
    def connect(
        self,
        hostname='transfer.rc.hms.harvard.edu',
        username='rh183',
        password='',
        port=22,
        key_filename=None,
        look_for_keys=True,
    ):
        """
        Connect to the remote server.
        Args:
            hostname (str):
                Hostname of the remote server.
            username (str):
                Username to log in with.
            password (str):
                Password to log in with.
                Is not stored.
            port (int):
                Port to connect to.
                sftp is always port 22.
        """
        self.client.connect(
            hostname=hostname,
            username=username,
            password=password,
            port=port, 
            key_filename=key_filename,
            look_for_keys=look_for_keys, 
            allow_agent=look_for_keys,
        )
        self.ssh = self.client.invoke_shell()

    def send(self, cmd='ls', append_enter=True):
        """
        Send a command to the remote server.
        Args:
            cmd (str):
                Command to send.
            append_enter (bool):
                Whether or not to append an enter
                 (backslash n) to the command.
                This is usually necessary to send.
        """
        if append_enter:
            cmd += '\n'
        self.ssh.send(cmd)
    
    def receive(self, timeout=None, verbose=True, throw_error=False):
        """
        Receive data from the remote server.
        Args:
            timeout (int):
                Timeout for receiving data.
                If None, will use self.recv_timeout.
            verbose (bool):
                Whether or not to print progress.
        """
        if timeout is None:
            timeout = self.recv_timeout
        self.ssh.settimeout(timeout)
        
        try:
            out = self.ssh.recv(self.nbytes, ).decode('utf-8');
            if verbose:
                print(out)
        except:
            if throw_error:
                raise
            else:
                if verbose:
                    print('Timeout')
                out = None
        return out
    
    def send_receive(
        self, 
        cmd='ls', 
        append_enter=True,
        post_send_wait_t=0.1, 
        timeout=None, 
        verbose=None,
    ):
        """
        Send a command to the remote server, wait for 
         some time, and then receive the output.
        Args:
            cmd (str):
                Command to send.
            append_enter (bool):
                Whether or not to append an enter
                 (backslash n) to the command.
                This is usually necessary to send.
            post_send_wait_t (float):
                Time to wait after sending the command.
            timeout (int):
                Timeout for receiving data.
                If None, will use self.recv_timeout.
            verbose (bool):
                Whether or not to print progress.
        """
        self.send(cmd=cmd, append_enter=append_enter)
        time.sleep(post_send_wait_t)
        out = self.receive(timeout=timeout, verbose=verbose)   
        return out 
    
    def expect(
        self,
        str_success='(base) [rh183',
        partial_match=True,
        recv_timeout=0.3,
        total_timeout=60,
        sleep_time=0.1,
        error_on_timeout=False,
        verbose=None,
    ):
        """
        Wait for a string to appear in the output.
        Args:
            str_success (str):
                String to wait for.
            partial_match (bool):
                Whether or not to allow a partial match.
            recv_timeout (float):
                Timeout for receiving data per 
                 check iteration.
            total_timeout (float):
                Total time to wait for the string.
            sleep_time (float):
                Time to sleep between checks.
                Allows for keyboard interrupts.
            error_on_timeout (bool):
                Whether or not to raise an error if the
                total_timeout is reached.
            verbose (bool):
                Whether or not to print progress.
                0/False: no printing
                1/True: will print recv outputs.
                2: will print expect progress.
                None: will default to self.verbose (1 or 2).
        """
        t_start = time.time()
        
        if recv_timeout is None:
            recv_timeout = self.recv_timeout
        if verbose is None:
            verbose = self.verbose
            
        success = False
        out=''
        while success is False:
            if verbose==2:
                print(f'=== expecting, t={time.time() - t_start} ===')
            
            out = self.receive(timeout=recv_timeout, verbose=False)
            if out is None:
                if verbose==2:
                    print("expect: nothing received")
            elif verbose > 0:
                print(out)
                    
            if partial_match and (out is not None):
                if str_success in out:
                    success = True
            else:
                if str_success == out:
                    success = True
            
            time.sleep(sleep_time)

            if time.time() - t_start > total_timeout:
                if error_on_timeout:
                    raise TimeoutError(f'Expect timeout: {str_success} not found in {total_timeout} seconds')
                else:
                    break
        
        if verbose==2:
            if success:
                print(f'expect succeeded')
            else:
                print(f'expect failed')
                
        return out, success

    def send_expect(
        self,
        cmd='ls',
        str_success='(base) [rh183',
        partial_match=True,
        recv_timeout=0.3,
        total_timeout=60,
        sleep_time=0.1,
        error_on_timeout=False,
        verbose=None,
    ):
        """
        Send a command to the remote server, wait for 
         a string to appear in the output, and then 
         receive the output.
        Args:
            cmd (str):
                Command to send.
            str_success (str):
                String to wait for.
            partial_match (bool):
                Whether or not to allow a partial match.
            recv_timeout (float):
                Timeout for receiving data per 
                 check iteration.
            total_timeout (float):
                Total time to wait for the string.
            sleep_time (float):
                Time to sleep between checks.
                Allows for keyboard interrupts.
            error_on_timeout (bool):
                Whether or not to raise an error if the
                total_timeout is reached.
            verbose (bool):
                Whether or not to print progress.
                0/False: no printing
                1/True: will print recv outputs.
                2: will print expect progress.
                None: will default to self.verbose (1 or 2).
        """
        self.send(cmd=cmd)
        out, success = self.expect(
            str_success=str_success,
            partial_match=partial_match,
            recv_timeout=recv_timeout,
            total_timeout=total_timeout,
            sleep_time = sleep_time,
            error_on_timeout=error_on_timeout,
            verbose=verbose,
        )
        return out, success

    def initialize_sftp(self):
        """
        Initialize the SFTP client.
        """
        self.sftp = self.client.open_sftp()
        return self.sftp
        
    def close(self):
        self.ssh.close()
        
    def __del__(self):
        if self.ssh is not None:
            self.ssh.close()

            
    def o2_connect(
        self,
        hostname='transfer.rc.hms.harvard.edu',
        username='rh183',
        password='',
        skip_passcode=False,
        key_filename=None,
        look_for_keys=False,
        passcode_method=1,
        verbose=1,
    ):
        """
        Connect to the O2 cluster.
        Helper function with some hard-coded expectations.
        Args:
            hostname (str):
                Hostname of the remote server.
            username (str):
                Username to log in with.
            password (str):
                Password to log in with.
                Is not stored.
            passcode_method (int):
                Method to use for O2 passcode.
                1. Duo Push
                2. Phone call
                3. SMS passcodes
            verbose (int):
                0/False: no printing
                1/True: will print recv outputs.
                2: will print expect progress.
                None: will default to self.verbose (1 or 2).
            skip_passcode (bool):
                Whether or not to skip the passcode step.
        """
        self.connect(
            hostname=hostname,
            username=username,
            password=password,
            port=22,
            key_filename=key_filename,
            look_for_keys=look_for_keys,
        )
        
        if skip_passcode==False:
            self.expect(
                str_success=f'Passcode or option (1-3)',
                partial_match=True,
                recv_timeout=0.3,
                total_timeout=60,
                verbose=verbose,
            )
            
            self.send(cmd=str(passcode_method))
        
        self.expect(
            str_success=f'[{username}@',
            partial_match=True,
            recv_timeout=0.3,
            total_timeout=60,
            verbose=verbose,
        )

    def fasrc_connect(
        self,
        hostname='login.rc.fas.harvard.edu',
        username='rhakim',
        password=None,
        verbose=1,
    ):
        """
        Connect to the FASRC cluster using interactive authentication. \n
        Helper function with some hard-coded expectations.
        Args:
            hostname (str):
                Hostname of the remote server.
            username (str):
                Username to log in with.
            verbose (int):
                0/False: no printing
                1/True: will print recv outputs.
                2: will print expect progress. (Sets logging level to DEBUG)
                None: will default to self.verbose (1 or 2).
        """
        def interactive_handler(title, instructions, prompt_list):
            responses = []
            for prompt, echo in prompt_list:
                if "password" in prompt.lower():
                    if password is not None:
                        responses.append(password)
                    else:
                        ## Prompt user for password
                        responses.append(getpass.getpass(prompt))
                elif "verification" in prompt.lower():
                    responses.append(getpass.getpass(prompt))
                else:
                    print(f"Prompt not recognized: {prompt}")
            return responses
        
        if verbose > 1:
            import logging
            logging.basicConfig(level=logging.DEBUG)

        ## Below try statement is needed in order to set up the client attributes. 
        ## The easiest way is to just try to connect (and set attrs), fail and move on.
        try:
            self.client.connect(hostname=hostname, allow_agent=False, look_for_keys=False)
        except:
            pass
        
        transport = self.client.get_transport()
        assert transport is not None, "Failed to get transport from SSH client. Found `client.get_transport()` to be None."
        transport.auth_interactive(username=username, handler=interactive_handler)
        self.ssh = self.client.invoke_shell()
        self.expect(
            str_success=f'{username}@',
            partial_match=True,
            recv_timeout=0.3,
            total_timeout=60,
            verbose=verbose,
        )

class sftp_interface():
    """
    Interface to sftp with a remote server.
    Mostly a wrapper for paramiko.SFTPClient.
    Tested on O2 cluster at Harvard.
    RH 2022
    """
    def __init__(
        self,
        ssh_client=None,
        hostname="transfer.rc.hms.harvard.edu",
        port=22,
    ):
        """
        Args:
            ssh_obj (paramiko.SSHClient):
                SSHClient object to use.
                Can be taken from ssh_interface with
                 ssh_interface.client
            hostname (str):
                Hostname of the remote server.
            port (int):
                Port of the remote server.
        """
        paramiko = _require_paramiko()
        if ssh_client is None:
            self.transport = paramiko.Transport((hostname, port))  ## open a transport object
        else:
            if isinstance(ssh_client, ssh_interface):
                self.client = ssh_client
                self.sftp = self.client.open_sftp()
            elif isinstance(ssh_client, paramiko.SSHClient):
                self.client = ssh_client
                self.sftp = self.client.open_sftp()
            elif isinstance(ssh_client, paramiko.Channel):
                self.transport = ssh_client.get_transport()
                self.sftp = paramiko.SFTPClient.from_transport(self.transport)
                self.client = None
        
    def connect(
        self,
        username='rh183',
        password=''
    ):
        """
        Connect to the remote server.
        Args:
            username (str):
                Username to log in with.
            password (str):
                Password to log in with.
                Is not stored.
        """
        paramiko = _require_paramiko()
        if hasattr(self, 'transport'):
            self.transport.connect(None, username, password)  ## authorization
            self.sftp = paramiko.SFTPClient.from_transport(self.transport)  ## open sftp
        elif hasattr(self, 'client'):
            self.sftp = self.client.open_sftp()
        else:
            raise ValueError('No valid connection method found')
        
        self.ssh = ssh_interface(ssh_client=self.client)
    
    def put_dir(self, source, target, mkdir=True, verbose=True):
        '''
        Uploads the contents of the source directory to the target path.
        All subdirectories in source are created under target recusively.
        Args:
            source (str):
                Path to the source directory (local).
            target (str):
                Path to the target directory (remote).
            mkdir (bool):
                Make the target directory if it does not exist.
        '''
        source = Path(source).resolve()
        target = Path(target).resolve()
        
        if mkdir:
            self.mkdir_safe(str(target), ignore_existing=True)

        for item in os.listdir(source):
            if os.path.isfile(source / item):
                if verbose:
                    print(f'uploading {source / item}   to   {target / item}')
                self.sftp.put(str(source / item) , str(target / item))
            else:
                self.mkdir_safe(str(target / item) , ignore_existing=True)
                self.put_dir(source / item , target / item)

    def get_dir(self, source, target, mkdir=True, verbose=True, prog_bars=False):
        '''
        Downloads the contents of the source directory to the target path. All
        subdirectories in source are created under target recusively.
        
        Args:
            source (str):
                Path to the source directory (remote).
            target (str):
                Path to the target directory (local).
            mkdir (bool):
                Whether or not to create the target directory.
            verbose (bool):
                Whether or not to print progress of files being downloaded.
        '''
        if mkdir:
            Path(target).mkdir(parents=True, exist_ok=True)
        source = Path(source).resolve()
        target = Path(target).resolve()
        
        for item in self.sftp.listdir(str(source)):
            if self.isdir_remote(str(source / item)):
                (target / item).mkdir(parents=True, exist_ok=True)  ## probably not necessary anymore since done above
                self.get_dir(source / item , target / item)
            else:
                if verbose:
                    print(f'downloading {source / item}   to   {target / item}')
                self.get(
                    remotepath=str(source / item), 
                    localpath=str(target / item),
                    mkdirs=mkdir,
                    prefetch=False,
                    prog_bar=prog_bars,
                )
        
    def mkdir_safe(self, path_remote, mode=511, ignore_existing=False):
        '''
        Augments mkdir by adding an option to not fail if the folder exists.
        Will fail if the outer directory does not exist.
        Args:
            path_remote (str):
                Path to the remote directory.
            mode (int):
                Mode to set the directory to.   
            ignore_existing (bool):
                Whether or not to ignore existing folders.
                If True, will not fail if the folder already exists.
                If False, will fail if the folder already exists.
        '''
        try:
            self.sftp.mkdir(path_remote, mode)
        except IOError:
            if ignore_existing:
                pass
            else:
                raise
    
    def mkdir_p(self, dir_remote):
        """
        Change to this directory, recursively making new folders if needed.
        Returns True if any folders were created.
        Args:
            dir_remote (str):
                Path to the remote directory.
        """
        if dir_remote == '/':
            # absolute path so change directory to root
            self.sftp.chdir('/')
            return
        if dir_remote == '':
            # top-level relative directory must exist
            return
        try:
            self.sftp.chdir(dir_remote) # sub-directory exists
        except IOError:
            dirname, basename = os.path.split(dir_remote.rstrip('/'))
            self.mkdir_p(dirname) # make parent directories
            self.sftp.mkdir(basename) # sub-directory missing, so created it
            self.sftp.chdir(basename)
            return True
    
    def isdir_remote(self, path):
        """
        Checks if a remote path is a directory.
        Args:
            path (str):
                Path to the remote directory.
        """
        try:
            return stat.S_ISDIR(self.sftp.stat(path).st_mode)
        except IOError:
            #Path does not exist, so by definition not a directory
            return False

    def search_recursive(
        self, 
        dir_outer='.', 
        reMatch='', 
        reMatch_in_path='',
        depth=6,
        find_files=True,
        find_folders=True,
        natsorted=True,
        verbose=True
    ):
        """
        Searches a remote directory recursively for files or directories
         matching a pattern.
        Args:
            sftp (paramiko.SFTPClient):
                SFTPClient object.
            dir_outer (str):
                Current working directory.
            reMatch (str):
                Regular expression to search for.
            reMatch_in_path (str):
                Regular expression to search for in the path.
            depth (int):
                Maximum depth (number of hierarchical subdirectories) to search.
            find_files (bool):
                Whether or not to search for files.
            find_folders (bool):
                Whether or not to search for folders.
            natsorted (bool):
                Whether or not to return the results natsorted.
            verbose (bool):
                Whether or not to print the paths of the files found.
                If False, no output is printed.
                If True or 1, prints the paths of the files found.
                If >1, prints the current search directory.

        Returns:
            list:cod
                List of paths to the files found.
        """
        search_results = []

        def _recursive_search(search_results, sftp, cwd='.', depth_current=0, verbose=True):
            if depth_current > depth:
                return search_results
            contents = {name: stat.S_ISDIR(attr.st_mode)  for name, attr in zip(sftp.listdir(cwd), sftp.listdir_attr(cwd))}
            for name, isdir  in contents.items():
                if (isdir and find_folders) or (not isdir and find_files):
                    if re.search(reMatch, name) and re.search(reMatch_in_path, str(Path(cwd) / name)):
                        path_found = str(Path(cwd) / name)
                        search_results.append(path_found)
                        print(path_found) if verbose else None

                if isdir:
                    search_results = _recursive_search(
                        search_results=search_results,
                        sftp=sftp, 
                        cwd=str(Path(cwd) / name), 
                        depth_current=depth_current+1,
                        verbose=verbose
                    )

                print(f'cwd: {cwd}, name: {name}, isdir: {isdir}, depth: {depth_current}') if verbose > 1 else None
            return search_results
        
        if natsorted:
            natsort = _require_natsort()
            fn_sort = natsort.natsorted
        else:
            fn_sort = lambda x: x
        
        return fn_sort(_recursive_search(
            search_results, 
            self.sftp, 
            cwd=dir_outer, 
            depth_current=0, 
            verbose=verbose,
        ))
        
    def list_fileSizes_recursive(self, directory='.'):
        """
        Lists the sizes of all files in a remote directory recursively.
        Args:
            sftp (paramiko.SFTPClient):
                SFTPClient object.
            directory (str):
                directory to search within.

        Returns:
            dict_of_sizes:
                Dictionary of paths and sizes of files
                 {'path_of_file', size) of all files found.
        """
        sizes = []

        def _recursive_list_sizes(sftp, cwd='.', sizes=[]):
            contents = {name: stat.S_ISDIR(attr.st_mode)  for name, attr in zip(sftp.listdir(cwd), sftp.listdir_attr(cwd))}
            for name, isdir  in contents.items():
                if isdir:
                    sizes = _recursive_list_sizes(sftp, str(Path(cwd) / name), sizes=sizes)
                else:
                    size = sftp.stat(str(Path(cwd) / name)).st_size
                    sizes.append((str(Path(cwd) / name), size))
            return sizes

        return dict(_recursive_list_sizes(self.sftp, cwd=directory, sizes=sizes))

    
    def get_fileProperties(self, paths, error_on_missing=False):
        """
        Return a dictionary of properties for one or more files.

        Args:
            paths (str or list of str):
                Path or list of paths to the files.
            error_on_missing (bool):
                Whether or not to raise an error if a file is missing.
        Returns:
            dict:
                Dictionary of properties for each file.
        """
        if isinstance(paths, str):
            paths = [paths]
        
        props_raw = {}
        for path in paths:
            try:
                props_raw[path] = self.sftp.stat(path)
            except FileNotFoundError:
                if error_on_missing:
                    raise FileNotFoundError(f'File not found: {path}')
                else:
                    props_raw[path] = None

        props = {}
        for path, prop in props_raw.items():
            if prop is not None:
                props[path] = {
                    'size': prop.st_size,
                    'last_modified': datetime.datetime.fromtimestamp(prop.st_mtime).strftime('%Y-%m-%d %H:%M:%S'),
                    'last_accessed': datetime.datetime.fromtimestamp(prop.st_atime).strftime('%Y-%m-%d %H:%M:%S'),
                    'permissions': prop.st_mode,
                }
            else:
                props[path] = None
        
        if len(props) == 1:
            props = props[paths[0]]

        return props


    def hierarchical_folderSizes(self, directory='.'):
        """
        Makes a hierarchical dict of dicts containing the 
         directory tree structure and sizes of each folder.
        Args:
            sftp (paramiko.SFTPClient):
                SFTPClient object.
            directory (str):
                Current working directory.

        Returns:
            h_dict:
                Hierarchical dict of dicts with fields
                 {name: children, '_size': size} where children contains
                 either dictionaries for children.
        """

        def _recursive_sum_sizes(sftp, cwd='.', size=0):
            contents = {name: {} for name in sftp.listdir(cwd)}
            isdir = [stat.S_ISDIR(attr.st_mode) for attr in sftp.listdir_attr(cwd)]
            for (name,v), isd in zip(list(contents.items()), isdir):
                if isd==False:  ## file
                    contents[name]['_size'] = sftp.stat(str(Path(cwd) / name)).st_size
                else:                     ## directory
                    contents[name], contents[name]['_size'] = _recursive_sum_sizes(sftp, cwd=str(Path(cwd) / name), size=0)
                size += contents[name]['_size']
            return contents, size

        out = _recursive_sum_sizes(self.sftp, cwd=directory, size=0)
        ret = {directory: out[0]}
        ret[directory]['_size'] = out[1]
        return ret

    def exists(self, path):
        try:
            self.sftp.stat(path)
            return True
        except FileNotFoundError:
            return False

    def get(
        self, 
        remotepath, 
        localpath, 
        mkdirs=True,
        prefetch=False, 
        callback=None, 
        prog_bar=True,
    ):
        """
        Get a file from the remote host.
        Uses a fancy class to do the downloading. It doesn't slow
         down due to the size of the file.
        Some extra stuff in here like a progress bar.
        RH 2022

        Args:
            remote_path (str):
                Path to the file on the remote host.
            local_path (str):
                Path to the file on the local host.
            prefetch (bool):
                Whether or not to prefetch the file.
            callback (function):
                Callback function to call within self.sftp.get
            prog_bar (bool):
                Whether or not to show a progress bar.
        
        Returns:
            None
        """
        Path(localpath).parent.mkdir(parents=True, exist_ok=True) if mkdirs else None

        with _TqdmWrap(ascii=False, unit='b', unit_scale=True) as pbar:
            def conj_func(a,b):
                pbar.viewBar(a, b)
                if callback is not None:
                    callback(a, b)
                    
            with self.sftp.open(remotepath, 'rb') as f_in, open(localpath, 'wb') as f_out:
                _SFTPFileDownloader(
                    f_in=f_in,
                    f_out=f_out,
                    prefetch=prefetch,
                    callback=conj_func if prog_bar else callback,
                ).download()

    def compare_directories(
        self,
        dir_remote,
        dir_local,
        depth=None,
        verbose=True,
    ):
        """
        Compares a local directory with a remote directory using file metadata.

        Args:
            dir_remote (str):
                Path to the remote directory.
            dir_local (str):
                Path to the local directory.
            depth (int):
                Maximum depth to search. If None, will search recursively
                through all subdirectories and files. If 0, will only search the
                current directory.
            verbose (bool):
                Whether or not to print progress: \n
                    * 0/False: no printing \n
                    * 1/True: will print found files \n
                    * 2: will print the current directory being searched.
        """
        print(f"Walking remote directory: {dir_remote}") if verbose else None
        paths_rel_remote = [str(Path(p).relative_to(dir_remote)) for p in self.walk_remote(dir_remote, depth=depth, verbose=verbose > 1)]
        print(f"Walking local directory: {dir_local}") if verbose else None
        paths_rel_local = [str(Path(p).relative_to(dir_local)) for p in self.walk_local(dir_local, depth=depth, verbose=verbose > 1)]

        paths_common = set(paths_rel_remote) & set(paths_rel_local)
        paths_unique_remote = set(paths_rel_remote) - set(paths_rel_local)
        paths_unique_local = set(paths_rel_local) - set(paths_rel_remote)

        def _metadata_remote(path):
            stat = self.sftp.stat(str(Path(dir_remote) / path))
            return {
                'size': stat.st_size,
                'last_modified': datetime.datetime.fromtimestamp(stat.st_mtime).strftime('%Y-%m-%d %H:%M:%S'),
                'permissions': stat.st_mode,
            }
        def _metadata_local(path):
            stat = (Path(dir_local) / path).stat()
            return {
                'size': stat.st_size,
                'last_modified': datetime.datetime.fromtimestamp(stat.st_mtime).strftime('%Y-%m-%d %H:%M:%S'),
                'permissions': stat.st_mode,
            }

        metadata_remote = {path: _metadata_remote(path) for path in paths_common | paths_unique_remote}
        metadata_local = {path: _metadata_local(path) for path in paths_common | paths_unique_local}

        
        metadata_matched = {path: {meta: metadata_remote[path][meta] == metadata_local[path][meta] for meta in metadata_remote[path]} for path in paths_common}

        return {
            'paths_common': paths_common,
            'paths_unique_remote': paths_unique_remote,
            'paths_unique_local': paths_unique_local,
            'metadata_remote': metadata_remote,
            'metadata_local': metadata_local,
            'metadata_matched': metadata_matched,
        }

    def walk_remote(
        self,
        dir_remote,
        depth=None,
        verbose=True,
    ):
        """
        Walks through a remote directory. Yielding the path to each file. Does
        not return empty directories.
        Args:
            dir_remote (str):
                Path to the remote directory.
            depth (int):
                Maximum depth to search. If None, will search recursively
                through all subdirectories and files. If 0, will only search the
                current directory.
            verbose (bool):
                Whether or not to print progress: \n
                    * 0/False: no printing \n
                    * 1/True: will print found files \n
                    * 2: will print the current directory being searched.
        """
        if depth is None:
            depth = float('inf')
        
        def _recursive_walk(sftp, cwd='.', depth_current=0, verbose=True):
            if depth_current > depth:
                return
            contents = {name: stat.S_ISDIR(attr.st_mode)  for name, attr in zip(sftp.listdir(cwd), sftp.listdir_attr(cwd))}
            for name, isdir in contents.items():
                if isdir:
                    print(f'cwd: {cwd}, name: {name}, isdir: {isdir}, depth: {depth_current}') if verbose > 1 else None
                    yield from _recursive_walk(sftp, cwd=str(Path(cwd) / name), depth_current=depth_current+1, verbose=verbose)
                else:
                    print(f"found: {str(Path(cwd) / name)}") if verbose else None
                    yield str(Path(cwd) / name)
                    
        return _recursive_walk(self.sftp, cwd=dir_remote, depth_current=0, verbose=verbose)
    
    def walk_local(
        self,
        dir_local,
        depth=None,
        verbose=True,
    ):
        """
        Walks through a local directory. Yielding the path to each file. Does
        not return empty directories.
        Args:
            dir_local (str):
                Path to the local directory.
            depth (int):
                Maximum depth to search. If None, will search recursively
                through all subdirectories and files. If 0, will only search the
                current directory.
            verbose (bool):
                Whether or not to print progress: \n
                    * 0/False: no printing \n
                    * 1/True: will print found files \n
                    * 2: will print the current directory being searched.
        """
        if depth is None:
            depth = float('inf')
        
        def _recursive_walk(cwd, depth_current=0, verbose=True):
            if depth_current > depth:
                return
            contents = {name: Path(cwd / name).is_dir() for name in os.listdir(cwd)}
            for name, isdir in contents.items():
                if isdir:
                    print(f'cwd: {cwd}, name: {name}, isdir: {isdir}, depth: {depth_current}') if verbose > 1 else None
                    yield from _recursive_walk(cwd / name, depth_current=depth_current+1, verbose=verbose)
                else:
                    print(f"found: {cwd / name}") if verbose else None
                    yield str(cwd / name)
                    
        return _recursive_walk(Path(dir_local), depth_current=0, verbose=verbose)

    def close(self):
        if hasattr(self, 'transport'):
            self.transport.close()
        elif hasattr(self, 'client'):
            self.client.close()
        elif hasattr(self, 'sftp'):
            self.sftp.close()

    def refresh_connection(self):
        self.connect()

def make_rsync_command(
    source, 
    destination, 
    recursive=True, 
    preserve_attributes=True, 
    verbose=True, 
    dry_run=False,
    delete=False, 
    compress=False,
    archive=False,
    exclude=None, 
    include=None, 
    extra_options=None,
):
    """
    Constructs an rsync command based on the provided options.
    RH 2023

    Args:
        source (str): 
            The source path for the rsync command.
        destination (str): 
            The destination path for the rsync command.
        recursive (bool, optional): 
            If True, transfer directories recursively. This means that
            directories will be copied along with their contents.
        preserve_attributes (bool, optional): 
            If True, preserve file attributes. Like: permissions, symbolic
            links, etc.
        verbose (bool, optional): 
            If True, enable verbose output. 
        dry_run (bool, optional): 
            If True, perform a trial run without changes. 
        delete (bool, optional): 
            If True, delete files in the destination that are not in source.
            Ensures that the destination is an exact copy of the source.
        compress (bool, optional): 
            If True, compress file data during transfer. 
        archive (bool, optional): 
            If True, use archive mode (preserves permissions, symbolic links,
            etc.). 
        exclude (list[str], optional): 
            List of patterns to exclude. 
        include (list[str], optional): 
            List of patterns to include. .
        extra_options (list[str], optional): 
            Additional rsync options to include. 

    Returns:
        str: 
            The constructed rsync command.

    Example:
        ``` make_rsync_command("/path/to/source", "/path/to/dest", delete=True)
        "rsync -r -p -v --delete /path/to/source /path/to/dest" ```
    """

    # Initialize exclude and include if they're None
    if exclude is None:
        exclude = []
    if include is None:
        include = []

    # Start with the rsync command
    cmd = ["rsync"]
    
    # Include options based on function arguments
    if archive:
        cmd.append("-a")
    else:
        if recursive:
            cmd.append("-r")
        if preserve_attributes:
            cmd.append("-p")
    if verbose:
        cmd.append("-v")
    if dry_run:
        cmd.append("--dry-run")
    if delete:
        cmd.append("--delete")
    if compress:
        cmd.append("-z")
    for pattern in exclude:
        cmd.extend(["--exclude", pattern])
    for pattern in include:
        cmd.extend(["--include", pattern])
    if extra_options:
        cmd.extend(extra_options)
    
    # Add source and destination
    cmd.extend([source, destination])

    return ' '.join(cmd)


"""
Implemented by casperdcl here: https://github.com/tqdm/tqdm/issues/311#issuecomment-387066847
"""
try:
    from tqdm.auto import tqdm
except ImportError:
    class _TqdmWrap(object):
        # tqdm not installed - construct and return dummy/basic versions
        def __init__(self, *a, **k):
            pass

        def viewBar(self, a, b):
            # original version
            res = a / int(b) * 100
            sys.stdout.write('\rComplete precent: %.2f %%' % (res))
            sys.stdout.flush()

        def __enter__(self):
            return self

        def __exit__(self, *a):
            return False
else:
    class _TqdmWrap(tqdm):
        def viewBar(self, a, b):
            self.total = int(b)
            self.update(int(a - self.n))  # update pbar with increment


class _SFTPFileDownloader:
    """
    Helper class to download large file with paramiko sftp client with limited number of concurrent requests.
    Implemented by vznncv here: https://gist.github.com/vznncv/cb454c21d901438cc228916fbe6f070f
    """

    _DOWNLOAD_MAX_REQUESTS = 48
    _DOWNLOAD_MAX_CHUNK_SIZE = 0x8000

    def __init__(self, f_in: paramiko.SFTPFile, f_out: typing.BinaryIO, callback=None, prefetch=False):
        self._paramiko = _require_paramiko()
        self.f_in = f_in
        self.f_out = f_out
        self.callback = callback
        self.prefetch = prefetch

        self.requested_chunks = {}
        self.received_chunks = {}
        self.saved_exception = None

    def download(self):
        file_size = self.f_in.stat().st_size
        requested_size = 0
        received_size = 0

        while True:
            # send read requests
            while len(self.requested_chunks) + len(self.received_chunks) < self._DOWNLOAD_MAX_REQUESTS and \
                    requested_size < file_size:
                chunk_size = min(self._DOWNLOAD_MAX_CHUNK_SIZE, file_size - requested_size)
                request_id = self._sftp_async_read_request(
                    fileobj=self,
                    file_handle=self.f_in.handle,
                    offset=requested_size,
                    size=chunk_size
                )
                self.requested_chunks[request_id] = (requested_size, chunk_size)
                requested_size += chunk_size

            # receive blocks if they are available
            # note: the _async_response is invoked
            self.f_in.sftp._read_response()
            self._check_exception()

            # write received data to output stream
            while True:
                chunk = self.received_chunks.pop(received_size, None)
                if chunk is None:
                    break
                _, chunk_size, chunk_data = chunk
                self.f_out.write(chunk_data)
                if self.callback is not None:
                    self.callback(received_size, file_size)

                received_size += chunk_size

            # check transfer status
            if received_size >= file_size:
                break

            # check chunks queues
            if not self.requested_chunks and len(self.received_chunks) >= self._DOWNLOAD_MAX_REQUESTS:
                raise ValueError("SFTP communication error. The queue with requested file chunks is empty and"
                                 "the received chunks queue is full and cannot be consumed.")

        return received_size

    def _sftp_async_read_request(self, fileobj, file_handle, offset, size):
        sftp_client = self.f_in.sftp
        paramiko = self._paramiko

        with sftp_client._lock:
            num = sftp_client.request_number

            msg = paramiko.Message()
            msg.add_int(num)
            msg.add_string(file_handle)
            msg.add_int64(offset)
            msg.add_int(size)

            sftp_client._expecting[num] = fileobj
            sftp_client.request_number += 1

        sftp_client._send_packet(paramiko.sftp.CMD_READ, msg)
        return num

    def _async_response(self, t, msg, num):
        paramiko = self._paramiko
        if t == paramiko.sftp.CMD_STATUS:
            # save exception and re-raise it on next file operation
            try:
                self.f_in.sftp._convert_status(msg)
            except Exception as e:
                self.saved_exception = e
            return
        if t != paramiko.sftp.CMD_DATA:
            raise paramiko.SFTPError("Expected data")
        data = msg.get_string()

        chunk_data = self.requested_chunks.pop(num, None)
        if chunk_data is None:
            return

        # save chunk
        offset, size = chunk_data

        if size != len(data):
            raise paramiko.SFTPError(f"Invalid data block size. Expected {size} bytes, but it has {len(data)} size")
        self.received_chunks[offset] = (offset, size, data)

    def _check_exception(self):
        """if there's a saved exception, raise & clear it"""
        if self.saved_exception is not None:
            x = self.saved_exception
            self.saved_exception = None
            raise x


def pw_encode(pw):
    import base64
    if pw is not None:
        return base64.b64encode(pw.encode("utf-8"))
    else:
        return None
def pw_decode(pw):
    import base64
    if pw is not None:
        return base64.b64decode(pw).decode("utf-8")
    else:
        return None
        



from __future__ import annotations

from typing import Optional, Union, List, Tuple, Dict, Any, Iterable, Sequence

import urllib


def _require_sqlalchemy():
    try:
        import sqlalchemy
    except ImportError as e:
        raise ImportError(
            "sqlalchemy is required for SQL connection operations. "
            "Install with `pip install bnpm[sql]`."
        ) from e
    return sqlalchemy


def _require_pandas():
    try:
        import pandas as pd
    except ImportError as e:
        raise ImportError(
            "pandas is required for SQL dataframe operations. "
            "Install with `pip install bnpm[io]` or `pip install bnpm[sql]`."
        ) from e
    return pd


def make_url(
    sql_type: str = "mysql",
    username: str = "root",
    password: str = "password",
    host: str = "localhost",
    port: int = None,
    database: str = "",
    quote_password: bool = False,
) -> str:
    """
    Formats a URL for a SQL database connection.\n
    Works with MySQL, PostgreSQL, and SQLite.\n
    RH 2024

    Args:
        sql_type (str):
            Type of SQL database. Options are "mysql", "postgresql", or "sqlite"
        username (str):
            Username for the database
        password (str):
            Password for the database
        host (str):
            Host for the database
        port (int):
            Port for the database.\n
            If 'default', then the default port for the SQL type is used.\n
        database (str):
            Name of the database. Leave blank to connect to the server without a
            database.
        quote_password (bool):
            Use quote_plus to format the passwords with special characters like:
            !@#$%^&*(){}/[]|:;<>,.?\~`_+-=\n

    Returns:
        (str):
            URL for the database connection
    """
    if (port is not None) and (port != "default"):
        port = f":{port}"
    elif port is None:
        port = ""

    if quote_password:
        password = urllib.parse.quote_plus(password)

    if sql_type == "mysql":
        if port == "default":
            port = 3306
        return f"mysql+pymysql://{username}:{password}@{host}{port}/{database}"
    
    elif sql_type == "postgresql":
        if port == "default":
            port = 5432
        return f"postgresql+psycopg2://{username}:{password}@{host}{port}/{database}"
    
    elif sql_type == "sqlite":
        return f"sqlite:///{database}"
    
    else:
        raise ValueError("sql_type must be 'mysql', 'postgresql', or 'sqlite'")
    

def parse_url(
    url: str
) -> Dict[str, str]:
    """
    Parses / decomposes a URL for a SQL database connection into its components.
    \n
    RH 2024

    Args:
        url (str):
            URL for the database connection

    Returns:
        (dict):
            Dictionary of the decomposed URL
    """

    def assert_split_in_n(s: str, split: str, n=2):
        assert len(s.split(split)) == n, f"URL must have exactly {n-1} '{split}' characters"

    if "mysql" in url:
        sql_type = "mysql"
    elif "postgresql" in url:
        sql_type = "postgresql"
    elif "sqlite" in url:
        sql_type = "sqlite"
    else:
        raise ValueError("URL type not recognized")
    
    ## Find the :// or :/// and remove everything before it
    if sql_type == "mysql":
        assert_split_in_n(url, "://", n=2)
        url = url.split("://")[1]
    elif sql_type == "postgresql":
        assert_split_in_n(url, "://", n=2)
        url = url.split("://")[1]
    elif sql_type == "sqlite":
        assert_split_in_n(url, "://", n=2)
        url = url.split(":///")[1]
        
    ## Split the username and password
    username, password_host_port_database = url.split(":")[0], url.split(":")[1]

    ## Split the password, host, port, and database
    assert_split_in_n(password_host_port_database, "@", n=2)
    password, host_port_database = password_host_port_database.split("@")
    
    ## Split the host, port, and database
    ### Handle if there is no port and if there is no database
    if "/" in host_port_database:
        assert_split_in_n(host_port_database, "/", n=2)
        host_port, database = host_port_database.split("/")
    else:
        host_port, database = host_port_database, None
    
    ## Split the host and port
    if ":" in host_port:
        assert_split_in_n(host_port, ":", n=2)
        host, port = host_port.split(":")
    else:
        host, port = host_port, None

    return {
        "sql_type": sql_type,
        "username": username,
        "password": password,
        "host": host,
        "port": port,
        "database": database
    }


def make_sql_connection(
    url: str,
    echo: bool = False,
    pool_size: int = 5,
    max_overflow: int = 10,
    pool_recycle: int = 3600,
    pool_pre_ping: bool = True,
    **kwargs,
) -> sqlalchemy.engine.base.Connection:
    """
    Makes a connection to a SQL database.
    RH 2024

    Args:
        url (str):
            URL for the database connection
        echo (bool):
            Whether to echo SQL commands to the console
        pool_size (int):
            Number of connections to keep in the pool
        max_overflow (int):
            Maximum number of connections to create
        pool_recycle (int):
            Number of seconds after which to recycle a connection
        pool_pre_ping (bool):
            Whether to ping the server before using a connection
        **kwargs:
            Additional keyword arguments for the connection

    Returns:
        (sqlalchemy.engine.base.Connection):
            SQL connection object
    """
    sqlalchemy = _require_sqlalchemy()
    return sqlalchemy.create_engine(
        url=url,
        echo=echo,
        pool_size=pool_size,
        max_overflow=max_overflow,
        pool_recycle=pool_recycle,
        pool_pre_ping=pool_pre_ping,
        **kwargs
    ).connect()


def get_available_databases(
    url: str,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> dict:
    """
    Gets a list of all databases available on the server.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        kwargs_connection (dict):
            Additional keyword arguments to be passed to
            sqlalchemy.create_engine
    
    Returns:
        (pd.DataFrame):
            DataFrame of all databases
    """
    sqlalchemy = _require_sqlalchemy()
    pd = _require_pandas()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        query = "SHOW DATABASES"
    elif "postgresql" in str(engine):
        query = "SELECT datname FROM pg_database WHERE datistemplate = false"
    elif "sqlite" in str(engine):
        raise ValueError("SQLite does not support this operation")
    else:
        raise ValueError("Connection type not recognized")
    
    with engine.begin() as connection:
        return pd.read_sql(query, connection)


def get_available_tables(
    url: str,
    database: Optional[str] = None,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> pd.DataFrame:
    """
    Gets a list of all tables available in the database.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        database (str):
            Name of the database. If None, then the default database is used.
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    
    Returns:
        (pd.DataFrame):
            DataFrame of all tables
    """
    sqlalchemy = _require_sqlalchemy()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    return sqlalchemy.inspect(engine).get_table_names(schema=database)

def get_table_columns(
    url: str,
    table: str,
    database: Optional[str] = None,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> pd.DataFrame:
    """
    Gets a list of all columns in the table.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        table (str):
            Name of the table
        database (str):
            Name of the database. If None, then the default database is used.
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    
    Returns:
        (pd.DataFrame):
            DataFrame of all columns
    """
    sqlalchemy = _require_sqlalchemy()
    pd = _require_pandas()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        if database is None:
            raise ValueError("database must be specified for MySQL")
        query = f"SHOW COLUMNS FROM {database}.{table}"
    elif "postgresql" in str(engine):
        if database is None:
            query = f"SELECT column_name FROM information_schema.columns WHERE table_name = '{table}'"
        else:
            query = f"SELECT column_name FROM information_schema.columns WHERE table_name = '{table}' AND table_catalog = '{database}'"
    elif "sqlite" in str(engine):
        raise ValueError("SQLite does not support this operation")
    else:
        raise ValueError("Connection type not recognized")

    with engine.begin() as connection:
        return pd.read_sql(query, connection)


def get_table_data(
    url: str,
    table: str,
    database: Optional[str] = None,
    limit: Optional[int] = None,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> pd.DataFrame:
    """
    Gets a list of all data in the table.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        table (str):
            Name of the table
        database (str):
            Name of the database. If None, then the default database is used.
        limit (int):
            Number of rows to return. If None, then all rows are returned.
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    
    Returns:
        (pd.DataFrame):
            DataFrame of all data
    """
    sqlalchemy = _require_sqlalchemy()
    pd = _require_pandas()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        if database is None:
            raise ValueError("database must be specified for MySQL")
        query = f"SELECT * FROM {database}.{table}"
    elif "postgresql" in str(engine):
        if database is None:
            query = f"SELECT * FROM {table}"
        else:
            query = f"SELECT * FROM {database}.{table}"
    elif "sqlite" in str(engine):
        if database is not None:
            raise ValueError("database must be None for SQLite")
        query = f"SELECT * FROM {table}"
    else:
        raise ValueError("Connection type not recognized")
    
    if limit is not None:
        query += f" LIMIT {limit}"
        
    with engine.begin() as connection:
        return pd.read_sql(query, connection)


def create_database(
    url: str,
    database: str,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> None:
    """
    Creates a new database on the server using a context manager.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        database (str):
            Name of the database to create
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    """
    sqlalchemy = _require_sqlalchemy()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        query = f"CREATE DATABASE {database}"
    elif "postgresql" in str(engine):
        query = f"CREATE DATABASE {database}"
    elif "sqlite" in str(engine):
        raise ValueError("SQLite does not support this operation")
    else:
        raise ValueError("Connection type not recognized")
        
    with engine.begin() as connection:
        connection.execute(sqlalchemy.text(query))


def drop_database(
    url: str,
    database: str,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> None:
    """
    Drops a database from the server.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        database (str):
            Name of the database to drop
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    """
    sqlalchemy = _require_sqlalchemy()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        query = f"DROP DATABASE {database}"
    elif "postgresql" in str(engine):
        query = f"DROP DATABASE {database}"
    elif "sqlite" in str(engine):
        raise ValueError("SQLite does not support this operation")
    else:
        raise ValueError("Connection type not recognized")
        
    with engine.begin() as connection:
        connection.execute(sqlalchemy.text(query))


def create_table(
    url: str,
    table: str,
    columns: Dict[str, str],
    database: Optional[str] = None,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> None:
    """
    Creates a new table in the database.
    RH 2024

    Args:
        url (str):
            URL for the database connection
        table (str):
            Name of the table to create
        columns (dict):
            Dictionary of column names and types. Example: {"column1": "INT",
            "column2": "VARCHAR(255)"}
        database (str):
            Name of the database. If None, then the default database is used.
        kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
    """
    sqlalchemy = _require_sqlalchemy()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        if database is None:
            raise ValueError("database must be specified for MySQL")
        query = f"CREATE TABLE {database}.{table} ("
        for column, column_type in columns.items():
            query += f"{column} {column_type}, "
        query = query[:-2] + ")"
    elif "postgresql" in str(engine):
        if database is None:
            query = f"CREATE TABLE {table} ("
        else:
            query = f"CREATE TABLE {database}.{table} ("
        for column, column_type in columns.items():
            query += f"{column} {column_type}, "
        query = query[:-2] + ")"
    elif "sqlite" in str(engine):
        if database is not None:
            raise ValueError("database must be None for SQLite")
        query = f"CREATE TABLE {table} ("
        for column, column_type in columns.items():
            query += f"{column} {column_type}, "
        query = query[:-2] + ")"
    else:
        raise ValueError("Connection type not recognized")
        
    with engine.begin() as connection:
        connection.execute(sqlalchemy.text(query))


def drop_table(
    url: str,
    table: str,
    database: Optional[str] = None,
    kwargs_connection: Optional[Dict[str, Any]] = {},
) -> None:
    """
    Drops a table from the database.
    RH 2024
    
    Args:
        url (str):
            URL for the database connection
        table (str):
            Name of the table to drop
        database (str):
            Name of the database. If None, then the default database is used.
         kwargs_connection (dict):
            Additional keyword arguments to be passed to sqlalchemy.create_engine
   """
    sqlalchemy = _require_sqlalchemy()
    engine = sqlalchemy.create_engine(url, **kwargs_connection)
    if "mysql" in str(engine):
        if database is None:
            raise ValueError("database must be specified for MySQL")
        query = f"DROP TABLE {database}.{table}"
    elif "postgresql" in str(engine):
        if database is None:
            query = f"DROP TABLE {table}"
        else:
            query = f"DROP TABLE {database}.{table}"
    elif "sqlite" in str(engine):
        if database is not None:
            raise ValueError("database must be None for SQLite")
        query = f"DROP TABLE {table}"
    else:
        raise ValueError("Connection type not recognized")
        
    with engine.begin() as connection:
        connection.execute(sqlalchemy.text(query))


# def insert_data(
#     url: str,
#     table: str,
#     data: Union[Dict[str, Any], List[Dict[str, Any]]],
#     database: Optional[str] = None,
#     kwargs_connection: Optional[Dict[str, Any]] = {},
# ) -> None:
#     """
#     Inserts data into the table.
#     RH 2024
    
#     Args:
#         url (str):
#             URL for the database connection
#         table (str):
#             Name of the table to insert data into
#         data (dict or list of dict):
#             Data to insert. If a single dictionary, then a single row is inserted.
#             If a list of dictionaries, then multiple rows are inserted.
#         database (str):
#             Name of the database. If None, then the default database is used.
#         kwargs_connection (dict):
#             Additional keyword arguments to be passed to sqlalchemy.create_engine
#     """
#     engine = sqlalchemy.create_engine(url, **kwargs_connection)
#     if "mysql" in str(engine):
#         if database is None:
#             raise ValueError("database must be specified for MySQL")
#         if isinstance(data, dict):
#             query = f"INSERT INTO {database}.{table} ({', '.join(data.keys())}) VALUES ({', '.join([str(value) for value in data.values()])})"
#         elif isinstance(data, list):
#             query = f"INSERT INTO {database}.{table} ({', '.join(data[0].keys())}) VALUES "
#             for row in data:
#                 query += f"({', '.join([str(value) for value in row.values()])}), "
#             query = query[:-2]
#     elif "postgresql" in str(engine):
#         if database is None:
#             if isinstance(data, dict):
#                 query = f"INSERT INTO {table} ({', '.join(data.keys())}) VALUES ({', '.join([str(value) for value in data.values()])})"
#             elif isinstance(data, list):
#                 query = f"INSERT INTO {table} ({', '.join(data[0].keys())}) VALUES "
#                 for row in data:
#                     query += f"({', '.join([str(value) for value in row.values()])}), "
#                 query = query[:-2]
#         else:
#             if isinstance(data, dict):
#                 query = f"INSERT INTO {database}.{table} ({', '.join(data.keys())}) VALUES ({', '.join([str(value) for value in data.values()])})"
#             elif
